==> untahris-2.0-src.tgz: untahris/tetris.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// UTetris.

#include <math.h>

struct OTetris : TPlayObject {
  public:
  int ix;
  int iy;
  int bx;
  int by;
  bool brick[4][4];
  int n; // blocks in brick

  int tBlockingType(int x, int y) {
    if(g.map[y][x] == cFree) return 2;
    CellType m = bigType(x, y);
    if(isBonus(m)) return 2;
    if(isSnakePart(g.map[y][x]) || g.map[y][x] == cTetris) return 1;
    return 0;
    }
  
  int tLiningType(int x, int y) {
    // 0: not forming lines, falling yes
    // 1: not forming lines, falling no
    // 2: limit for forming lines
    // 3: forms lines
    CellType m = g.map[y][x];
    if(m == cFree) return 0;
    if(isSnakePart(m) || m == cPermaWall)
      return 2;
    if(m == cTetrisDown || m == cSnakeTRON || m == cBreakWall)
      return 3;
    return 1;
    }

  int blockFits() {
    int bf = 2;
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
      int bf2 = tBlockingType(bx+px, by+py);
      if(bf2 < bf) bf = bf2;
      }
    return bf;
    }
  
  void putBlock(CellType ct, int x, int y) {
    CellType m = bigType(x, y);
    if(isBonus(m)) {
      switch(m) {
        case cBonusLevel:
          d.playSoundAt(d.sndBlue, x, y);
          ow()->score += scoreGain() * 10;
          ow()->level += 10;
          break;
        case cBonusScore:
          d.playSoundAt(d.sndGreen, x, y);
          ow()->score += scoreGain() * 2;
          break;
        case cBonusAmmo:
          d.playSoundAt(d.sndRed, x, y);
          ow()->score += scoreGain();
          ow()->ammo += 10;
        }
          
      placeBigRandomly(m, 0);
      clearBig(x, y, cFree, 0);
      }
    g.mapSet(x,y, ct, owner);
    }
  
  void tryLines(int x, int y) {
    int x0 = x, x1 = x;
    int t0 = 3, t1 = 3;
    while(t0 == 3) {x0--; t0 = tLiningType(x0, y);}
    while(t1 == 3) {x1++; t1 = tLiningType(x1, y);}
    if((t0 == 2 && t1 == 2) || (x1-x0 > 16)) {
      int len = x1-x0-1;
      ow()->level += len;
      ow()->score += scoreGain() * len;
      for(int x2=x0+1; x2<x1; x2++) {
        int y2 = y;
        while(1) {
          int t = tLiningType(x2, y2-1);
          if(t == 1 || t == 2) {
            g.mapSet(x2,y2, cFree, 0);
            break;
            }
          else {
            g.mapSet(x2,y2, g.map[y2-1][x2], g.owner[y2-1][x2]);
            y2--;
            }
          }
        }
      }
    }

  void putBlock(CellType ct) {
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px])
      putBlock(ct, bx+px, by+py);
    }
  
  void tryLines() {
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px])
      tryLines(bx+px, by+py);
    }
  
  void takeBlock() {
    n = 0;
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
      int y = by+py, x = bx+px;
      if(g.map[by+py][bx+px] == cTetris && g.owner[by+py][bx+px] == owner) {
        g.mapSet(bx+px, by+py, cFree, 0);
        n++;
        }
      else brick[py][px] = false;
      }
    }
  
  bool createNewBrick() {
    char* bricks[7] = {
      "..#...#...#...#.",
      ".....##..##.....",
      ".....#...##..#..",
      "......#..##..#..",
      ".....#...##...#.",
      ".....##..#...#..",
      ".....##...#...#."
      };
    bx = ix;
    by = iy;
    int btype = g.rand() % 7;
    for(int b=0; b<16; b++) brick[b/4][b%4] = bricks[btype][b] == '#';
    int bf = blockFits();
    switch(bf) {
      case 2:
        putBlock(cTetris);
        n = 4;
        break;
      case 0:
        g.deleteObject(this);
        ow()->typeval = -1;
        break;
      // just wait in case of 1
      }
    }
  
  OTetris(int _owner, int _x, int _y) : ix(_x), iy(_y) {
    owner = _owner;
    time = g.time;
    g.objects.push_back(this);
    n = 0;
    }
  
  OTetris(TBuffer &b) {
    b.get(owner); 
    b.get(ix); b.get(iy);
    b.get(bx); b.get(by);
    b.get(time); b.get(n);
    for(int ky=0; ky<4; ky++)
    for(int kx=0; kx<4; kx++) b.get(brick[ky][kx]);
    }
  void save(TBuffer &b) {
    b.putChar(4);
    b.put(owner); 
    b.put(ix); b.put(iy);
    b.put(bx); b.put(by);
    b.put(time); b.put(n);
    for(int ky=0; ky<4; ky++)
    for(int kx=0; kx<4; kx++) b.put(brick[ky][kx]);
    }

  void go() {
    if(n) takeBlock();
    if(n) {
      by++;
      int bf = blockFits();
      if(bf == 2)
        putBlock(cTetris);
      else if(bf == 1) {
        by--;
        putBlock(cTetris);
        }
      else {
        by--;
        d.playSoundAt(d.sndPlace, bx+1, by+1);
        putBlock(cTetrisDown);
        tryLines();
        n = 0;
        }
      }
    else { 
      createNewBrick();
      }
    time += spd() * 5;
    }

  void ref1() {
    for(int x=0; x<4; x++) for(int y=0; y<2; y++)
      swap(brick[y][x], brick[3-y][x]);
    }
  
  void ref2() {
    for(int x=0; x<4; x++) for(int y=0; y<x; y++)
      swap(brick[y][x], brick[x][y]);
    }
  
  void reactToKey(PlayerKey pk, bool pressed) {
    if(!pressed) return;
    if(n) switch(pk) {
      case keyLeft:
        takeBlock();
        bx--;
        if(blockFits() != 2) bx++;
        putBlock(cTetris);
        break;
      case keyRight:
        takeBlock();
        bx++;
        if(blockFits() != 2) bx--;
        putBlock(cTetris);
        break;
      case keyDown:
        takeBlock();
        by++;
        if(blockFits() != 2) by--;
        putBlock(cTetris);
        break;
      case keyFire: case keyWater:
        takeBlock();
        while(blockFits()) by++;
        by--;
        putBlock(cTetris);
        break;
      case keyUp:
        takeBlock();
        ref1(); ref2();
        if(blockFits() != 2) {ref2(); ref1();}
        putBlock(cTetris);
        break;
      case keyEarth:
        if(ow()->ammo >= 2)
          d.playSoundAt(d.sndShoot, bx+1, by+1);
        for(int py=0; py<4; py++) for(int px=0; px<4; px++)
          if(brick[py][px] && (py == 3 || !brick[py+1][px]) && ow()->useAmmo(2))
            new OBullet(owner, Ctr(bx+px), Ctr(by+py) + 2048, 0, bulletSpeed());
        break;
      case keyAir:
        if(ow()->ammo >= 2)
          d.playSoundAt(d.sndShoot, bx+1, by+1);
        for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
          if((px == 3 || !brick[py][px+1]) && ow()->useAmmo(1))
            new OBullet(owner, Ctr(bx+px) + 2048, Ctr(by+py), bulletSpeed(), 0);
          if((px == 0 || !brick[py][px-1]) && ow()->useAmmo(1))
            new OBullet(owner, Ctr(bx+px) - 2048, Ctr(by+py), -bulletSpeed(), 0);
          }
        break;
      }
    }
  };

==> untahris-3.1-src.tgz: untahris/tetris.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// UTetris.

#include <math.h>

struct OTetris : TPlayObject {
  public:
  int32 ix;
  int32 iy;
  int32 bx;
  int32 by;
  bool brick[4][4];
  int32 n; // blocks in brick

  int tBlockingType(int x, int y) {
    if(g.map[y][x] == cFree) return 2;
    if(g.map[y][x] == cXunux) return 2;
    CellType m = bigType(x, y);
    if(isBonus(m)) return 2;
    if(isSnakePart(g.map[y][x]) || g.map[y][x] == cTetris) return 1;
    return 0;
    }
  
  int tLiningType(int x, int y) {
    // 0: not forming lines, falling yes
    // 1: not forming lines, falling no
    // 2: limit for forming lines
    // 3: forms lines
    CellType m = g.map[y][x];
    if(m == cFree) return 0;
    if(isSnakePart(m) || m == cPermaWall)
      return 2;
    if(m == cTetrisDown || m == cSnakeTRON || m == cBreakWall)
      return 3;
    return 1;
    }

  int blockFits() {
    int bf = 2;
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
      int bf2 = tBlockingType(bx+px, by+py);
      if(bf2 < bf) bf = bf2;
      }
    return bf;
    }
  
  void putBlock(CellType ct, int x, int y) {
    CellType m = bigType(x, y);
    if(isBonus(m)) {
      switch(m) {
        case cBonusLevel:
          d.playSoundAt(d.sndBlue, x, y);
          ow()->score += scoreGain() * 10;
          ow()->level += 10;
          break;
        case cBonusScore:
          d.playSoundAt(d.sndGreen, x, y);
          ow()->score += scoreGain() * 2;
          break;
        case cBonusAmmo:
          d.playSoundAt(d.sndRed, x, y);
          ow()->score += scoreGain();
          ow()->gainAmmo();
          break;
        }
          
      if(gc.regenerate(m)) 
        placeBigRandomly(m, 0);
      clearBig(x, y, cFree, 0);
      }

    if(g.map[y][x] == cXunux)
      destroyAt(x, y, owner, 0);

    g.mapSet(x,y, ct, owner);
    }
  
  void tryLines(int x, int y) {
    int x0 = x, x1 = x;
    int t0 = 3, t1 = 3;
    while(t0 == 3) {x0--; t0 = tLiningType(x0, y);}
    while(t1 == 3) {x1++; t1 = tLiningType(x1, y);}
    if((t0 == 2 && t1 == 2) || (x1-x0 > gc.utetrisMinLength + 1)) {
      int len = x1-x0-1;
      ow()->level += len;
      ow()->score += scoreGain() * len;
      for(int x2=x0+1; x2<x1; x2++) {
        int y2 = y;
        while(1) {
          int t = tLiningType(x2, y2-1);
          if(t == 1 || t == 2) {
            g.mapSet(x2,y2, cFree, 0);
            break;
            }
          else {
            g.mapSet(x2,y2, g.map[y2-1][x2], g.owner[y2-1][x2]);
            y2--;
            }
          }
        }
      }
    }

  void putBlock(CellType ct) {
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px])
      putBlock(ct, bx+px, by+py);
    }
  
  void tryLines() {
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px])
      tryLines(bx+px, by+py);
    }
  
  void takeBlock() {
    n = 0;
    for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
      int y = by+py, x = bx+px;
      if(g.map[by+py][bx+px] == cTetris && g.owner[by+py][bx+px] == owner) {
        g.mapSet(bx+px, by+py, cFree, 0);
        n++;
        }
      else brick[py][px] = false;
      }
    }
  
  bool createNewBrick() {
    char* bricks[7] = {
      "..#...#...#...#.",
      ".....##..##.....",
      ".....#...##..#..",
      "......#..##..#..",
      ".....#...##...#.",
      ".....##..#...#..",
      ".....##...#...#."
      };
    bx = ix;
    by = iy;
    int btype = g.rand() % 7;
    for(int b=0; b<16; b++) brick[b/4][b%4] = bricks[btype][b] == '#';
    int bf = blockFits();
    switch(bf) {
      case 2:
        putBlock(cTetris);
        n = 4;
        break;
      case 0:
        g.deleteObject(this);
        ow()->typeval = -1;
        break;
      // just wait in case of 1
      }
    }
  
  OTetris(int _owner, int _x, int _y) : ix(_x), iy(_y) {
    owner = _owner;
    time = g.time;
    g.objects.push_back(this);
    n = 0;
    }
  
  OTetris(TBuffer &b) {
    b.get(owner); 
    b.get(ix); b.get(iy);
    b.get(bx); b.get(by);
    b.get(time); b.get(n);
    for(int ky=0; ky<4; ky++)
    for(int kx=0; kx<4; kx++) b.get(brick[ky][kx]);
    }
  void save(TBuffer &b) {
    b.putChar(4);
    b.put(owner); 
    b.put(ix); b.put(iy);
    b.put(bx); b.put(by);
    b.put(time); b.put(n);
    for(int ky=0; ky<4; ky++)
    for(int kx=0; kx<4; kx++) b.put(brick[ky][kx]);
    }

  void go() {
    if(n) takeBlock();
    if(n) {
      by++;
      int bf = blockFits();
      if(bf == 2)
        putBlock(cTetris);
      else if(bf == 1) {
        by--;
        putBlock(cTetris);
        }
      else {
        by--;
        d.playSoundAt(d.sndPlace, bx+1, by+1);
        putBlock(cTetrisDown);
        tryLines();
        n = 0;
        }
      }
    else { 
      createNewBrick();
      }
    time += spd() * 5;
    }

  void ref1() {
    for(int x=0; x<4; x++) for(int y=0; y<2; y++)
      swap(brick[y][x], brick[3-y][x]);
    }
  
  void ref2() {
    for(int x=0; x<4; x++) for(int y=0; y<x; y++)
      swap(brick[y][x], brick[x][y]);
    }
  
  void reactToKey(PlayerKey pk, bool pressed) {
    if(!pressed) return;
    if(n) switch(pk) {
      case keyLeft:
        takeBlock();
        bx--;
        if(blockFits() != 2) bx++;
        putBlock(cTetris);
        break;
      case keyRight:
        takeBlock();
        bx++;
        if(blockFits() != 2) bx--;
        putBlock(cTetris);
        break;
      case keyDown:
        takeBlock();
        by++;
        if(blockFits() != 2) by--;
        putBlock(cTetris);
        break;
      case keyFire: case keyWater:
        takeBlock();
        while(blockFits()) by++;
        by--;
        putBlock(cTetris);
        break;
      case keyUp:
        takeBlock();
        ref1(); ref2();
        if(blockFits() != 2) {ref2(); ref1();}
        putBlock(cTetris);
        break;
      case keyEarth:
        if(ow()->ammo >= ec().ammoPerShoot)
          d.playSoundAt(d.sndShoot, bx+1, by+1);
        for(int py=0; py<4; py++) for(int px=0; px<4; px++)
          if(brick[py][px] && (py == 3 || !brick[py+1][px]) && ow()->useAmmo())
            new OBullet(owner, Ctr(bx+px), Ctr(by+py) + 2048, 0, bulletSpeed());
        break;
      case keyAir:
        if(ow()->ammo >= ec().ammoPerShoot)
          d.playSoundAt(d.sndShoot, bx+1, by+1);
        for(int py=0; py<4; py++) for(int px=0; px<4; px++) if(brick[py][px]) {
          if((px == 3 || !brick[py][px+1]) && ow()->useAmmo((ec().ammoPerShoot+1)/2))
            new OBullet(owner, Ctr(bx+px) + 2048, Ctr(by+py), bulletSpeed(), 0);
          if((px == 0 || !brick[py][px-1]) && ow()->useAmmo((ec().ammoPerShoot+1)/2))
            new OBullet(owner, Ctr(bx+px) - 2048, Ctr(by+py), -bulletSpeed(), 0);
          }
        break;
      }
    }
  };
